package org.rcisoft.business.cmsActivity.service.impl;

import cn.hutool.core.lang.Assert;
import org.apache.commons.lang3.StringUtils;
import org.rcisoft.business.cmsActivity.dao.CmsActivityRepository;
import org.rcisoft.business.cmsActivity.entity.CmsActivity;
import org.rcisoft.business.memInfo.dao.MemInfoRepository;
import org.rcisoft.core.service.CyRedisService;
import org.rcisoft.core.util.CyUserUtil;
import org.rcisoft.core.util.IStockCallback;
import org.redisson.api.RLock;
import org.redisson.api.RedissonClient;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.stereotype.Service;
import redis.clients.jedis.Jedis;
import redis.clients.jedis.JedisPool;
import redis.clients.jedis.JedisPoolConfig;

import javax.annotation.Resource;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.TimeUnit;

/**
 * 扣库存
 *
 */
@Service
public class StockService {
    Logger logger = LoggerFactory.getLogger(StockService.class);

    @Autowired
    private RedissonClient redissonClient;

    /**
     * 不限库存
     */
    public static final long UNINITIALIZED_STOCK = -3L;

    /**
     * Redis 客户端
     */
    @Autowired
    private RedisTemplate<String, Object> redisTemplate;

    @Autowired(required = false)
    private CyRedisService cyRedisServiceImpl;

    @Autowired
    private MemInfoRepository memInfoRepository;
    @Autowired
    private CmsActivityRepository activityRepository;
    /**
     * 执行扣库存的脚本
     */
    public static final String STOCK_LUA;
    /**
     * 执行扣库存和男生人数的脚本
     */
    public static final String STOCK_LUA2;
    /**
     * 执行扣库存和女生人数的脚本
     */
    public static final String STOCK_LUA3;


    @Value("${spring.data.redis.host}")
    private String redisIP;

    @Value("${spring.data.redis.port}")
    private String redisPort;

    @Value("${spring.data.redis.password}")
    private String password;

    @Value("${spring.data.redis.database}")
    private int database;

    @Value("${spring.data.redis.timeout}")
    private int timeout;

    static {
        /**
         *
         * @desc 扣减库存Lua脚本
         * 库存（stock）-1：表示不限库存
         * 库存（stock）0：表示没有库存
         * 库存（stock）大于0：表示剩余库存
         *
         * @params 库存key
         * @return
         *   -3:库存未初始化
         *   -2:库存不足
         *   -1:不限库存
         *   大于等于0:剩余库存（扣减之后剩余的库存）
         *      redis缓存的库存(value)是-1表示不限库存，直接返回1
         */
        StringBuilder sb = new StringBuilder();
        sb.append("if (redis.call('exists', KEYS[1]) == 1) then");
        sb.append("    local stock = tonumber(redis.call('get', KEYS[1]));");
        sb.append("    local num = tonumber(ARGV[1]);");
        sb.append("    if (stock == -1) then");
        sb.append("        return -1;");
        sb.append("    end;");
        sb.append("    if (stock >= num) then");
        sb.append("        return redis.call('incrby', KEYS[1], 0 - num);");
        sb.append("    end;");
        sb.append("    return -2;");
        sb.append("end;");
        sb.append("return -3;");
        STOCK_LUA = sb.toString();

        /**
         *
         * @desc 扣减库存Lua脚本
         * 库存（stock）-1：表示不限库存
         * 库存（stock）0：表示没有库存
         * 库存（stock）大于0：表示剩余库存
         *
         * @params 库存key
         * @return
         *   -3:库存未初始化
         *   -2:库存不足
         *   -1:不限库存
         *   大于等于0:剩余库存（扣减之后剩余的库存）
         *      redis缓存的库存(value)是-1表示不限库存，直接返回1
         */
        StringBuilder sb2 = new StringBuilder();
        sb2.append("if (redis.call('exists', KEYS[1]) == 1 and redis.call('exists', KEYS[2]) == 1) then");
        sb2.append("    local totalStock = tonumber(redis.call('get', KEYS[1]));");
        sb2.append("    local maleStock = tonumber(redis.call('get', KEYS[2]));");
        sb2.append("    local num = tonumber(ARGV[1]);");
        sb2.append("    if (totalStock == -1) then");
        sb2.append("        return -1;");
        sb2.append("    end;");
        sb2.append("    if (maleStock == -1) then");
        sb2.append("        return -1;");
        sb2.append("    end;");
        sb2.append("    if (totalStock >= num and maleStock >= num) then");
        sb2.append("        redis.call('incrby', KEYS[2], 0 - num);"); // 扣减男生人数
        sb2.append("        return redis.call('incrby', KEYS[1], 0 - num);"); // 扣减总人数
        sb2.append("    end;");
        sb2.append("    return -2;"); // 库存不足
        sb2.append("end;");
        sb2.append("return -3;"); // 库存未初始化
        STOCK_LUA2 = sb2.toString();

        /**
         *
         * @desc 扣减库存Lua脚本
         * 库存（stock）-1：表示不限库存
         * 库存（stock）0：表示没有库存
         * 库存（stock）大于0：表示剩余库存
         *
         * @params 库存key
         * @return
         *   -3:库存未初始化
         *   -2:库存不足
         *   -1:不限库存
         *   大于等于0:剩余库存（扣减之后剩余的库存）
         *      redis缓存的库存(value)是-1表示不限库存，直接返回1
         */
        StringBuilder sb3 = new StringBuilder();
        sb3.append("if (redis.call('exists', KEYS[1]) == 1 and redis.call('exists', KEYS[2]) == 1) then");
        sb3.append("    local totalStock = tonumber(redis.call('get', KEYS[1]));");
        sb3.append("    local maleStock = tonumber(redis.call('get', KEYS[2]));");
        sb3.append("    local num = tonumber(ARGV[1]);");
        sb3.append("    if (totalStock == -1) then");
        sb3.append("        return -1;");
        sb3.append("    end;");
        sb3.append("    if (maleStock == -1) then");
        sb3.append("        return -1;");
        sb3.append("    end;");
        sb3.append("    if (totalStock >= num and maleStock >= num) then");
        sb3.append("        redis.call('incrby', KEYS[2], 0 - num);"); // 扣减女生人数
        sb3.append("        return redis.call('incrby', KEYS[1], 0 - num);"); // 扣减总人数
        sb3.append("    end;");
        sb3.append("    return -2;"); // 库存不足
        sb3.append("end;");
        sb3.append("return -3;"); // 库存未初始化
        STOCK_LUA3 = sb3.toString();
    }

    /**
     * @param key           库存key
     * @param expire        库存有效时间,单位秒
     * @param num           扣减数量
     * @param stockCallback 初始化库存回调函数
     * @return -2:库存不足; -1:不限库存; 大于等于0:扣减库存之后的剩余库存
     */
    public long stock(String key1, String key2, String key3, long expire, int num, IStockCallback stockCallback) {
        //扣总人数库存
        long stock1 = stock(key1, key2, key3, num);
        // 初始化库存
        if (stock1 == UNINITIALIZED_STOCK) {
            RLock redisLock = redissonClient.getLock("reduce_activity_count");
            try {
                // 获取锁
                if (redisLock.tryLock()) {
                    // 双重验证，避免并发时重复回源到数据库
                    stock1 = stock(key1, key2,key3,num);
                    if (stock1 == UNINITIALIZED_STOCK) {
                        // 获取初始化库存
                        final int initStock = stockCallback.getStock();
                        // 将库存设置到redis
                        redisTemplate.opsForValue().set(key1, initStock, expire, TimeUnit.SECONDS);
                        // 将库存的男生人数设置到redis
                        redisTemplate.opsForValue().set(key2, initStock, expire, TimeUnit.SECONDS);
                        // 将库存的女生人数设置到redis
                        redisTemplate.opsForValue().set(key3, initStock, expire, TimeUnit.SECONDS);
                        // 调一次扣库存的操作
                        stock1 = stock(key1, key2,key3,num);
                    }
                }
            } catch (Exception e) {
                logger.error(e.getMessage(), e);
            } finally {
                redisLock.unlock();
            }
        }
        return stock1;
    }

    /**
     * 加库存(还原库存)
     *
     * @param key    库存key
     * @param num    库存数量
     * @return
     */
    public long addStock(String key, int num) {

        return addStock(key, null, num);
    }

    /**
     * 加库存
     *
     * @param key    库存key
     * @param expire 过期时间（秒）
     * @param num    库存数量
     * @return
     */
    public long addStock(String key, Long expire, int num) {
        Object value = cyRedisServiceImpl.get(key);
        // 判断key是否存在，存在就直接更新
        if (value != null) {
            redisTemplate.opsForValue().increment(key, num);
            return num;
        }
//
//        if (expire != null) {
//            Assert.notNull(expire, "初始化库存失败，库存过期时间不能为null");
//            RLock redisLock = redissonClient.getLock(key);
//            try {
//                if (redisLock.tryLock()) {
//                    // 获取到锁后再次判断一下是否有key
//                    value = cyRedisServiceImpl.get(key)
//                    if (hasKey == null) {
//                        // 初始化库存
//                        redisTemplate.opsForValue().set(key, num, expire, TimeUnit.SECONDS);
//                    }
//                }
//            } catch (Exception e) {
//                logger.error(e.getMessage(), e);
//            } finally {
//                redisLock.unlock();
//            }
//        }

        return num;
    }

    /**
     * 获取库存
     *
     * @param key 库存key
     * @return -1:不限库存; 大于等于0:剩余库存
     */
    public int getStock(String key) {
        Integer stock = (Integer) redisTemplate.opsForValue().get(key);
        return stock == null ? -1 : stock;
    }

    /**
     * 扣库存
     *
     * @param key1 库存key
     * @param key2 库存男生人数key
     * @param key3 库存女生人数key
     * @param num 扣减库存数量
     * @return 扣减之后剩余的库存【-3:库存未初始化; -2:库存不足; -1:不限库存; 大于等于0:扣减库存之后的剩余库存】
     */
    private Long stock(String key1, String key2, String key3, int num) {
        // 脚本里的KEYS参数
        List<String> keys = new ArrayList<>();
        keys.add(key1);
        List<String> keys2 = new ArrayList<>();
        keys2.add(key1);
        keys2.add(key2);
        List<String> keys3 = new ArrayList<>();
        keys3.add(key1);
        keys3.add(key3);
        // 脚本里的ARGV参数
        List<String> args = new ArrayList<>();
        args.add(Integer.toString(num));
        long result = UNINITIALIZED_STOCK;
        JedisPoolConfig config = new JedisPoolConfig();
        try (JedisPool jedisPool = new JedisPool(config,redisIP,Integer.parseInt(redisPort), timeout,
                (StringUtils.isAnyEmpty(password) ? null : password),database);){
//        try (JedisPool jedisPool = new JedisPool(redisIP, Integer.parseInt(redisPort));){
            Jedis jedis = jedisPool.getResource();
            //获取报名人性别 截取key1的最后一个:后的值
            String goodsId = key1.substring(key1.lastIndexOf(":") + 1);
            CmsActivity activityInfo = activityRepository.selectById(Integer.valueOf(goodsId));
            String memSex = memInfoRepository.getMemSexByUserId(Integer.valueOf(CyUserUtil.getAuthenBusinessId()));
            //扣减库存总人数
            if (activityInfo.getMaleCount() == 0 && activityInfo.getFemaleCount() == 0){
                //活动的男女生人数都为0 不扣男女生人数
                result = (long) jedis.evalsha(jedis.scriptLoad(STOCK_LUA), keys , args);
            } else {
                if (memSex.equals("0") && activityInfo.getMaleCount() != 0){
                    //报名人为男生 且男生限制人数不为0  扣除总人数和男生人数
                    result = (long) jedis.evalsha(jedis.scriptLoad(STOCK_LUA2), keys2 , args);
                } else if (memSex.equals("1") && activityInfo.getFemaleCount() != 0){
                    //报名人为女生 且女生限制人数不为0  扣除总人数和女生人数
                    result = (long) jedis.evalsha(jedis.scriptLoad(STOCK_LUA3), keys3 , args);
                } else {
                    //只扣除总人数
                    result = (long) jedis.evalsha(jedis.scriptLoad(STOCK_LUA), keys , args);
                }
            }

        } catch (Exception e){
            logger.error(e.getMessage(),e);
        }
        logger.info("扣减之后剩余的库存：{}" , result);
        return result;
    }

}
